#include "udpweather.h"

#include <malloc.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>

namespace meteo {

UDPWeather::UDPWeather (std::string in_hostname, int in_hostport)
{
	hostname = std::string(in_hostname);
	hostport = in_hostport;
	sock = -1;
	conn_state = NOT_INITIAL;
}

UDPWeather::~UDPWeather (void)
{
	closeSocket ();
}

// create new connection..
int
UDPWeather::init (void)
{
	struct sockaddr_in bind_addr;
	int ret;
	struct hostent *server_info;

	if (conn_state == BAD_HOST)
		return -1;
	sock = socket (PF_INET, SOCK_DGRAM, 0);
	if (sock < 0) {
		mdebug(LOG_ERR, MDEBUG_LOG, 0, "cannot create meteo connection");
		return -1;
	}
	ret = fcntl (sock, F_SETFL, O_NONBLOCK);
	if (ret) {
		mdebug(LOG_ERR, MDEBUG_LOG, 0, "cannot set nonblock socket");
		return -1;
	}
	bind_addr.sin_family = AF_INET;
	bind_addr.sin_port = htons (0);
	bind_addr.sin_addr.s_addr = htonl (INADDR_ANY);

	serv_addr.sin_family = AF_INET;
	serv_addr.sin_port = htons (hostport);
  
	server_info = gethostbyname (hostname.c_str());
	if (!server_info) {
		mdebug(LOG_ERR, MDEBUG_LOG, 0, "cannot get ip for %s: %s", hostname.c_str(), strerror(errno));
		conn_state = BAD_HOST;
		return -1;
	}
	serv_addr.sin_addr = *(struct in_addr *) server_info->h_addr;
	ret = bind (sock, (struct sockaddr *) &bind_addr, sizeof(bind_addr));
	if (ret < 0) {
		mdebug(LOG_ERR, MDEBUG_LOG, 0, "cannot bind socket for %s, port %i: %s", hostname.c_str(), hostport, strerror(errno));
		conn_state = NOT_INITIAL;
		return -1;
	}
	return 0;
}

void
UDPWeather::closeSocket (void)
{
	if (sock >= 0)
		close (sock);
	sock = -1;
}

void
UDPWeather::receive (void)
{
	struct sockaddr_in	*other_addr;
	socklen_t		size = sizeof (other_addr);
	char			buf[4];
	int			ret;
	if (sock != -1) {
		ret = recvfrom(sock, buf, 3, 0, (struct sockaddr*) &other_addr, &size);
		if (ret < 0) {
			mdebug(LOG_DEBUG, MDEBUG_LOG, 0, "cannot receive data from socket, closing it");
			closeSocket ();
			return;
		}
		buf[ret] = '\0';
		if (strcmp (buf, "Ack")) {
			mdebug(LOG_ERR, MDEBUG_LOG, 0, "don't get Ack, get: '%s'", buf);
			closeSocket ();
		}
	}
}

void
UDPWeather::send (meteo::Station *s, time_t lastup)
{
	int		ret;
	char		*msg;
	struct tm	*tv;
	if (sock != -1) {
		receive ();
	}
	if (sock == -1) {
		ret = init ();
		if (ret == -1)
			return;
	}
	tv = gmtime (&lastup);

	asprintf (&msg, "rtWindSpeed=%f, rtInsideHum=%f, rtWindDir=%f, rtRainRate=%f, rtIsRaining=%i, rtWetness=%f, rtOutsideHum=%f, "
          "rtBaroCurr=%f, rtWindAvgSpeed=%f, rtOutsideTemp=%f, rtCloudTop=%f, rtCloudBottom=%f",
		s->getSensorValue(11,62), s->getSensorValue(10, 10), 
		s->getSensorValue(11,61), s->getSensorValue(11, 52), 
                (s->getSensorValue(11, 52) == 0 && s->getSensorValue (13, 20) == 0) ? 0 : 1,
                s->getSensorValue(13, 20), s->getSensorValue (11, 10),
                s->getSensorValue(10, 30), s->getSensorValue (11, 60),
                s->getSensorValue(11, 0), s->getSensorValue (13, 0),
		s->getSensorValue(14, 0));
	ret = sendto(sock, msg, strlen (msg), 0, (struct sockaddr*) &serv_addr, sizeof(serv_addr)); 

	free(msg);
}

}
